﻿using System;
using System.Text;
using System.Runtime.InteropServices;
using System.Collections;
using System.Collections.Specialized;
using System.Net.Sockets;
using System.IO;
using System.Runtime.Serialization;
using System.Collections.Generic;

/// <summary>
/// nMtpUtil 名前空間
/// </summary>
namespace nMtp
{
    /// <summary>
    /// ファイル情報構造体
    /// </summary>
	[StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
	public struct WIN32_FIND_DATA
	{
		/// <summary>
		/// 属性
		/// </summary>
		public uint dwFileAttributes;
		/// <summary>
		/// 作成日時
		/// </summary>
		public System.Runtime.InteropServices.ComTypes.FILETIME ftCreationTime;
		/// <summary>
		/// 最終アクセス日時
		/// </summary>
		public System.Runtime.InteropServices.ComTypes.FILETIME ftLastAccessTime;
		/// <summary>
		/// 最終更新日時
		/// </summary>
		public System.Runtime.InteropServices.ComTypes.FILETIME ftLastWriteTime;
		/// <summary>
		/// ファイルサイズ（上位32bit）
		/// </summary>
		public uint nFileSizeHigh;
		/// <summary>
		/// ファイルサイズ（下位32bit）
		/// </summary>
		public uint nFileSizeLow;
		/// <summary>
		/// 予約
		/// </summary>
		public uint dwReserved0;
		/// <summary>
		/// 予約
		/// </summary>
		public uint dwReserved1;
		/// <summary>
		/// ファイル名
		/// </summary>
		[MarshalAs(UnmanagedType.ByValTStr, SizeConst = 260)]
		public string cFileName;
		/// <summary>
		/// <para>8.3形式のファイル名</para>
		/// <para>MTPデバイス上のファイルまたはフォルダの場合 Object ID</para>
		/// </summary>
		[MarshalAs(UnmanagedType.ByValTStr, SizeConst = 14)]
		public string cAlternateFileName;
	}

	/// <summary>
	/// ファイルシステム基本クラス
	/// </summary>
	[SerializableAttribute]
	[ComVisibleAttribute(true)]
	public class FileSystemInfo : System.IO.FileSystemInfo {
		protected bool _read_flag = false;
		protected FileAttributes _attributes = (FileAttributes)(-1);
		protected DateTime _creation_time = new DateTime();
		protected DateTime _last_access_time = new DateTime();
		protected DateTime _last_write_time = new DateTime();
		protected long _length = 0;

        protected int MAX_PATH = 260;
		protected const uint FILE_ATTRIBUTE_DIRECTORY = 0x00000010;

		/// <summary>
		/// <c>FileSystemInfo</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		public FileSystemInfo() {
			OriginalPath = "";
			FullPath = "";
			_read_flag = false;
		}

		/// <summary>
		/// <c>FileSystemInfo</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		/// <param name="path">ファイルまたはフォルダ名</param>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	フォルダ が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		public FileSystemInfo(string path) {
			Directory.CheckInvalidPathName(path, false);

			OriginalPath = path;
			if(!NMtpCheckDevicePathName(path)) {
				FullPath = Path.GetFullPath(path);
			} else {
				FullPath = path;
			}
			_read_flag = false;
		}

		/// <summary>
		/// <c>FileSystemInfo</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		/// <param name="path">ファイルまたはフォルダ名</param>
		/// <param name="find_data">ファイルまたはフォルダの情報</param>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	フォルダ が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		public FileSystemInfo(string path, WIN32_FIND_DATA find_data)
        {
			Directory.CheckInvalidPathName(path, false);

			_attributes = (FileAttributes)find_data.dwFileAttributes;
			_creation_time = DateTime.FromFileTime((((long)find_data.ftCreationTime.dwHighDateTime) << 32) + find_data.ftCreationTime.dwLowDateTime).ToLocalTime();
			_last_access_time = DateTime.FromFileTime((((long)find_data.ftLastAccessTime.dwHighDateTime) << 32) + find_data.ftLastAccessTime.dwLowDateTime).ToLocalTime();
			_last_write_time = DateTime.FromFileTime((((long)find_data.ftLastWriteTime.dwHighDateTime) << 32) + find_data.ftLastWriteTime.dwLowDateTime).ToLocalTime();
			_length = ((long)find_data.nFileSizeHigh << 32) | find_data.nFileSizeLow;
			OriginalPath = Path.Combine(path, find_data.cFileName);
			FullPath = OriginalPath;
			_read_flag = true;
		}

		protected void GetStatus()
		{
			if(!_read_flag) {
				WIN32_FIND_DATA find_data;

				if(NMtpGetStatus(FullPath, out find_data)) {
					_read_flag = true;
					_attributes = (FileAttributes)find_data.dwFileAttributes;
					_creation_time = DateTime.FromFileTime((((long)find_data.ftCreationTime.dwHighDateTime) << 32) + find_data.ftCreationTime.dwLowDateTime);
					_last_access_time = DateTime.FromFileTime((((long)find_data.ftLastAccessTime.dwHighDateTime) << 32) + find_data.ftLastAccessTime.dwLowDateTime);
					_last_write_time = DateTime.FromFileTime((((long)find_data.ftLastWriteTime.dwHighDateTime) << 32) + find_data.ftLastWriteTime.dwLowDateTime);
					_length = ((long)find_data.nFileSizeHigh << 32) + find_data.nFileSizeLow;
				}
			}
		}

		protected static bool CheckDirectory(uint attr)
		{
			if((attr & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY) {
				return true;
			}
			return false;
		}

		protected static bool IsDots(string name)
		{
			if(name == "." || name == "..") {
				return true;
			}
			return false;
        }

		/// <summary>
		/// ファイルを削除
		/// </summary>
		public override void Delete() {
			if(IsDirectory) {
				Directory.Delete(FullPath);
			} else {
				File.Delete(FullPath);
			}
		}

		/// <summary>
		/// ファイル情報を更新
		/// </summary>
		public new void Refresh()
		{
			_read_flag = false;
			GetStatus();
		}

		/// <summary>
		/// ファイルが存在するかを示す値を取得
		/// </summary>
		public override bool Exists {
			get {
				if(IsDirectory) {
					return Directory.Exists(FullPath);
				} else {
					return File.Exists(FullPath);
				}
			}
		}

		/// <summary>
		/// ファイルの絶対パス名を取得
		/// </summary>
		public override string FullName {
			get {
				return FullPath;
			}
		}

		/// <summary>
		/// ファイルの名前を取得
		/// </summary>
		public override string Name {
			get {
				return Path.GetFileName(FullName);
			}
		}

		/// <summary>
		/// MTP デバイス上のファイルまたはフォルダかどうかを示す値を取得
		/// </summary>
		public bool IsMtpDevice {
			get {
				return NMtpCheckDevicePathName(FullPath);
			}
		}

		/// <summary>
		/// フォルダかどうかを示す値を取得
		/// </summary>
		public bool IsDirectory {
			get {
				GetStatus();
				if((_attributes & FileAttributes.Directory) == FileAttributes.Directory) {
					return true;
				}
				return false;
			}
		}

		/// <summary>
		/// <para>ファイルまたはフォルダの属性の取得または設定</para>
		/// <para>フォルダまたは MTP デバイス上のファイルの場合は設定できない</para>
		/// </summary>
		public virtual new FileAttributes Attributes {
			set {
				if(!IsMtpDevice) {
					if(!IsDirectory) {
						_attributes = value;
						System.IO.File.SetAttributes(FullPath, _attributes);
					}
				}
			}
			get {
				GetStatus();
				return _attributes;
			}
		}

		/// <summary>
		/// ファイルまたはフォルダの作成日時が有効かどうかを示す値を取得
		/// </summary>
		public bool IsValidCreationTime {
			get {
				if(_creation_time.ToLocalTime().CompareTo(DateTime.FromFileTime(0)) != 0) {
					return true;
				}
				return false;
			}
		}

		/// <summary>
		/// <para>ファイルまたはフォルダの作成日時(世界協定時刻)の取得または設定</para>
		/// <para>MTP デバイス上のファイルまたはフォルダの場合は設定できない</para>
		/// </summary>
		public virtual new DateTime CreationTimeUtc {
			set {
				if(!IsMtpDevice) {
					_creation_time = value;
					if(IsDirectory) {
						System.IO.Directory.SetCreationTimeUtc(FullPath, _creation_time);
					} else {
						System.IO.File.SetCreationTimeUtc(FullPath, _creation_time);
					}
				}
			}
			get {
				GetStatus();
				return _creation_time;
			}
		}
		/// <summary>
		/// <para>ファイルまたはフォルダの作成日時の取得または設定</para>
		/// <para>MTP デバイス上のファイルまたはフォルダの場合は設定できない</para>
		/// </summary>
		public virtual new DateTime CreationTime {
			set {
				if(!IsMtpDevice) {
					if(IsDirectory) {
						_creation_time = value.ToUniversalTime();
						System.IO.Directory.SetCreationTimeUtc(FullPath, _creation_time);
					} else {
						System.IO.File.SetCreationTimeUtc(FullPath, _creation_time);
					}
				}
			}
			get {
				GetStatus();
				return _creation_time.ToLocalTime();
			}
		}

		/// <summary>
		/// ファイルまたはフォルダの最終アクセス日時が有効かどうかを示す値を取得
		/// </summary>
		public bool IsValidLastAccessTime {
			get {
				if(_last_access_time.ToLocalTime().CompareTo(DateTime.FromFileTime(0)) != 0) {
					return true;
				}
				return false;
			}
		}
		/// <summary>
		/// <para>ファイルまたはフォルダの最終アクセス日時(世界協定時刻)の取得または設定</para>
		/// <para>MTP デバイス上のファイルまたはフォルダの場合は設定できない</para>
		/// </summary>
		public virtual new DateTime LastAccessTimeUtc {
			set {
				if(!IsMtpDevice) {
					_last_access_time = value;
					if(IsDirectory) {
						System.IO.Directory.SetLastAccessTimeUtc(FullPath, _last_access_time);
					} else {
						System.IO.File.SetLastAccessTimeUtc(FullPath, _last_access_time);
					}
				}
			}
			get {
				GetStatus();
				return _last_access_time;
			}
		}
		/// <summary>
		/// <para>ファイルまたはフォルダの最終アクセス日時の取得または設定</para>
		/// <para>MTP デバイス上のファイルまたはフォルダの場合は設定できない</para>
		/// </summary>
		public virtual new DateTime LastAccessTime {
			set {
				if(!IsMtpDevice) {
					_last_access_time = value.ToUniversalTime();
					if(IsDirectory) {
						System.IO.Directory.SetLastAccessTimeUtc(FullPath, _last_access_time);
					} else {
						System.IO.File.SetLastAccessTimeUtc(FullPath, _last_access_time);
					}
				}
			}
			get {
				GetStatus();
				return _last_access_time.ToLocalTime();
			}
		}
		/// <summary>
		/// ファイルまたはフォルダの最終書込日時が有効かどうかを示す値を取得
		/// </summary>
		public bool IsValidLastWriteTime {
			get {
				if(_last_write_time.ToLocalTime().CompareTo(DateTime.FromFileTime(0)) != 0) {
					return true;
				}
				return false;
			}
		}
		/// <summary>
		/// <para>ファイルまたはフォルダの最終書込日時(世界協定時刻)の取得または設定</para>
		/// <para>MTP デバイス上のファイルまたはフォルダの場合は設定できない</para>
		/// </summary>
		public virtual new DateTime LastWriteTimeUtc {
			set {
				if(!IsMtpDevice) {
					_last_write_time = value;
					if(IsDirectory) {
						System.IO.Directory.SetLastWriteTimeUtc(FullPath, _last_write_time);
					} else {
						System.IO.File.SetLastWriteTimeUtc(FullPath, _last_write_time);
					}
				}
			}
			get {
				GetStatus();
				return _last_write_time;
			}
		}
		/// <summary>
		/// <para>ファイルまたはフォルダの最終書込日時の取得または設定</para>
		/// <para>MTP デバイス上のファイルまたはフォルダの場合は設定できない</para>
		/// </summary>
		public virtual new DateTime LastWriteTime {
			set {
				if(!IsMtpDevice) {
					_last_write_time = value.ToUniversalTime();
					if(IsDirectory) {
						System.IO.Directory.SetLastWriteTime(FullPath, _last_write_time);
					} else {
						System.IO.File.SetLastWriteTimeUtc(FullPath, _last_write_time);
					}
				}
			}
			get {
				GetStatus();
				return _last_write_time.ToLocalTime();
			}
		}

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpCheckDevicePathName", CharSet=CharSet.Unicode)]
		protected static extern bool NMtpCheckDevicePathName(string path);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpGetStatus", CharSet=CharSet.Unicode, SetLastError=true)]
		protected static extern bool NMtpGetStatus(string path, out WIN32_FIND_DATA find_data);
	}

	/// <summary>
	/// フォルダ操作クラス
	/// </summary>
	[SerializableAttribute]
	[ComVisibleAttribute(true)]
	public class DirectoryInfo : FileSystemInfo {
		/// <summary>
		/// <para>コピー先にファイルがあれば失敗</para>
		/// <para>Copy/CopyFirst の flag で指定</para>
		/// </summary>
		public const int CopyFailExists = 0x00000001;
		/// <summary>
		/// <para>新しいファイルのみコピー</para>
		/// <para>Copy/CopyFirst の flag で指定</para>
		/// </summary>
		public const int CopyNewer = 0x00000010;
		/// <summary>
		/// <para>サブフォルダの中身もコピー</para>
		/// <para>Copy/CopyFirst の flag で指定</para>
		/// </summary>
		public const int CopySubFolder = 0x00000100;
		/// <summary>
		/// <para>一度にコピー</para>
		/// <para>CopyFirst の flag で指定</para>
		/// </summary>
		public const int CopyAll = 0x00001000;
		/// <summary>
		/// <para>一度にコピーして結果も保存しない</para>
		/// </summary>
		public const int CopyAllNoResult = 0x00002000;

		private const int INVALID_HANDLE_VALUE = -1;
		private const int MAX_DIR_PATH = 248;

		private IntPtr _handle = IntPtr.Zero;
		private string _source_file = "";
		private string _dest_file = "";
		private int _copy_result = 0;
		private int _copy_error = 0;

		private const int ERROR_FILE_NOT_FOUND = 2;
		private const int ERROR_PATH_NOT_FOUND = 3;
		private const int ERROR_FILE_EXISTS = 80;

		/// <summary>
		/// <c>DirectoryInfo</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		/// <param name="path">フォルダ名</param>
		public DirectoryInfo(string path) : base(path) {
		}

		/// <summary>
		/// <c>DirectoryInfo</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		/// <param name="path">フォルダ名</param>
		/// <param name="find_data">フォルダの情報</param>
		public DirectoryInfo(string path, WIN32_FIND_DATA find_data) : base(path, find_data)
        {
		}

		/// <summary>
		/// フォルダ作成
		/// </summary>
		///	<exception cref="IOException">
		///	フォルダを作成できない
		/// </exception>
		public void Create() {
			Directory.CreateDirectory(FullPath);
		}

		/// <summary>
		/// サブフォルダ作成
		/// </summary>
		/// <param name="path">サブフォルダ名</param>
		///	<exception cref="IOException">
		///	サブフォルダを作成できない
		/// </exception>
		public void CreateSubdirectory(string path) {
			Directory.CreateDirectory(Path.Combine(FullPath, path));
		}

		/// <summary>
		/// フォルダ削除
		/// </summary>
		///	<exception cref="IOException">
		///	サブフォルダを削除できない
		/// </exception>
		public override void Delete() {
			Directory.Delete(FullPath);
		}

		/// <summary>
		/// フォルダ削除
		/// </summary>
		/// <param name="recursive">true でサブフォルダ、ファイルも含めて削除</param>
		///	<exception cref="IOException">
		///	サブフォルダを削除できない
		/// </exception>
		public void Delete(bool recursive) {
			Directory.Delete(FullPath, recursive);
		}

		protected void GetDirectoryList(string path, string searchPattern, ref List<DirectoryInfo> list)
		{
            IntPtr h;
            WIN32_FIND_DATA find_data;
			h = NMtpFindFirstFile(Path.Combine(path, "*"), out find_data);
			if(h != (IntPtr)INVALID_HANDLE_VALUE) {
				do {
					if(CheckDirectory(find_data.dwFileAttributes)) {
						string name = find_data.cFileName.ToString();
						if(!IsDots(name)) {
							if(NMtpWildcardMatch(searchPattern, name)) {
								list.Add(new DirectoryInfo(path, find_data));
							}
							GetDirectoryList(Path.Combine(path, name), searchPattern, ref list);
						}
					}
				} while(NMtpFindNextFile(h, out find_data));
				NMtpFindClose(h);
			}
		}

		/// <summary>
		/// 指定されたフォルダ内のサブフォルダを取得
		/// </summary>
		/// <returns>指定したパス内のサブフォルダの配列。フォルダが見つからない場合は空の配列</returns>
		public DirectoryInfo [] GetDirectories()
		{
			return GetDirectories("*");
		}

		/// <summary>
		/// 指定されたフォルダ内のサブフォルダを取得
		/// </summary>
		/// <param name="searchPattern">検索文字列</param>
		/// <returns>指定したパス内のサブフォルダの配列。フォルダが見つからない場合は空の配列</returns>
		public DirectoryInfo [] GetDirectories(string searchPattern)
		{
			IntPtr h;
			WIN32_FIND_DATA find_data;
			List<DirectoryInfo> list = new List<DirectoryInfo>();
			h = NMtpFindFirstFile(Path.Combine(FullPath, searchPattern), out find_data);
			if(h != (IntPtr)INVALID_HANDLE_VALUE) {
				do {
					if(CheckDirectory(find_data.dwFileAttributes)) {
						list.Add(new DirectoryInfo(FullPath, find_data));
					}
				} while(NMtpFindNextFile(h, out find_data));
				NMtpFindClose(h);
			}
			return list.ToArray();
		}

		/// <summary>
		/// 指定されたフォルダ内のサブフォルダを取得
		/// </summary>
		/// <param name="searchPattern">検索文字列</param>
		/// <param name="searchOption">検索オプション</param>
		/// <returns>指定したパス内のサブフォルダの配列。フォルダが見つからない場合は空の配列</returns>
		public DirectoryInfo [] GetDirectories(string searchPattern, SearchOption searchOption)
		{
			if(searchOption == SearchOption.AllDirectories) {
				List<DirectoryInfo> list = new List<DirectoryInfo>();
				GetDirectoryList(FullPath, searchPattern, ref list);
				return list.ToArray();
			}
			return GetDirectories(searchPattern);
		}

		protected void GetFileList(string path, string searchPattern, ref List<FileInfo> list)
		{
            IntPtr h;
            WIN32_FIND_DATA find_data;
			h = NMtpFindFirstFile(Path.Combine(path, "*"), out find_data);
			if(h != (IntPtr)INVALID_HANDLE_VALUE) {
				do {
					string name = find_data.cFileName.ToString();
					if(CheckDirectory(find_data.dwFileAttributes)) {
						if(!IsDots(name)) {
							GetFileList(Path.Combine(path, name), searchPattern, ref list);
						}
					} else {
						if(NMtpWildcardMatch(searchPattern, name)) {
							list.Add(new FileInfo(path, find_data));
						}
					}
				} while(NMtpFindNextFile(h, out find_data));
				NMtpFindClose(h);
			}
		}

		/// <summary>
		/// 指定されたフォルダ内のファイルを取得
		/// </summary>
		/// <returns>指定したパス内のファイルの配列。ファイルが見つからない場合は空の配列</returns>
		public FileInfo [] GetFiles()
		{
			return GetFiles("*");
		}

		/// <summary>
		/// 指定されたフォルダ内のファイルを取得
		/// </summary>
		/// <param name="searchPattern">検索文字列</param>
		/// <returns>指定したパス内のファイルの配列。ファイルが見つからない場合は空の配列</returns>
		public FileInfo [] GetFiles(string searchPattern)
		{
			IntPtr h;
			WIN32_FIND_DATA find_data;
			List<FileInfo> list = new List<FileInfo>();

			h = NMtpFindFirstFile(Path.Combine(FullPath, searchPattern), out find_data);
			if(h != (IntPtr)INVALID_HANDLE_VALUE) {
				do {
					if(!CheckDirectory(find_data.dwFileAttributes)) {
						list.Add(new FileInfo(FullPath, find_data));
					}
				} while(NMtpFindNextFile(h, out find_data));
				NMtpFindClose(h);
			}
			return list.ToArray();
		}

		/// <summary>
		/// 指定されたフォルダ内のファイルを取得
		/// </summary>
		/// <param name="searchPattern">検索文字列</param>
		/// <param name="searchOption">検索オプション</param>
		/// <returns>指定したパス内のファイルの配列。ファイルが見つからない場合は空の配列</returns>
		public FileInfo [] GetFiles(string searchPattern, SearchOption searchOption)
		{
			if(searchOption == SearchOption.AllDirectories) {
				List<FileInfo> list = new List<FileInfo>();
				GetFileList(FullPath, searchPattern, ref list);
				return list.ToArray();
			}
			return GetFiles(searchPattern);
		}

		protected void GetFileSystemList(string path, string searchPattern, ref List<FileSystemInfo> list)
		{
            IntPtr h;
            WIN32_FIND_DATA find_data;
			h = NMtpFindFirstFile(Path.Combine(path, "*"), out find_data);
			if(h != (IntPtr)INVALID_HANDLE_VALUE) {
				do {
					string name = find_data.cFileName.ToString();
					if(CheckDirectory(find_data.dwFileAttributes)) {
						if(!IsDots(name)) {
							if(NMtpWildcardMatch(searchPattern, name)) {
								list.Add(new FileSystemInfo(path, find_data));
							}
							GetFileSystemList(Path.Combine(path, name), searchPattern, ref list);
						}
					} else {
						if(NMtpWildcardMatch(searchPattern, name)) {
							list.Add(new FileSystemInfo(path, find_data));
						}
					}
				} while(NMtpFindNextFile(h, out find_data));
				NMtpFindClose(h);
			}
		}

		/// <summary>
		/// 指定されたフォルダ内のファイルおよびサブフォルダを取得
		/// </summary>
		/// <returns>指定したパス内のサブフォルダのファイルおよびサブフォルダ名の配列。ファイルおよびサブフォルダが見つからない場合は空の配列</returns>
		public FileSystemInfo [] GetFileSystemInfos()
		{
			return GetFileSystemInfos("*");
		}

		/// <summary>
		/// 指定されたフォルダ内のファイルおよびサブフォルダを取得
		/// </summary>
		/// <param name="searchPattern">検索文字列</param>
		/// <returns>指定したパス内のサブフォルダのファイルおよびサブフォルダ名の配列。ファイルおよびサブフォルダが見つからない場合は空の配列</returns>
		public FileSystemInfo [] GetFileSystemInfos(string searchPattern)
		{
			IntPtr h;
			WIN32_FIND_DATA find_data;
			List<FileSystemInfo> list = new List<FileSystemInfo>();

			h = NMtpFindFirstFile(Path.Combine(FullPath, searchPattern), out find_data);
			if(h != (IntPtr)INVALID_HANDLE_VALUE) {
				do {
					if(!CheckDirectory(find_data.dwFileAttributes)) {
						list.Add(new FileSystemInfo(Path.Combine(FullPath, find_data.cFileName.ToString())));
					}
				} while(NMtpFindNextFile(h, out find_data));
				NMtpFindClose(h);
			}
			return list.ToArray();
		}

		/// <summary>
		/// 指定されたフォルダ内のファイルおよびサブフォルダを取得
		/// </summary>
		/// <param name="searchPattern">検索文字列</param>
		/// <param name="searchOption">検索オプション</param>
		/// <returns>指定したパス内のファイルおよびサブフォルダの配列。ファイルおよびサブフォルダが見つからない場合は空の配列</returns>
		public FileSystemInfo [] GetFileSystemInfos(string searchPattern, SearchOption searchOption)
		{
			if(searchOption == SearchOption.AllDirectories) {
				List<FileSystemInfo> list = new List<FileSystemInfo>();
				GetFileSystemList(FullPath, searchPattern, ref list);
				return list.ToArray();
			}
			return GetFiles(searchPattern);
		}

		/// <summary>
		/// フォルダファイルコピー
		/// </summary>
		/// <param name="destFolderName">コピー先フォルダ</param>
		/// <returns>true で成功</returns>
		public bool Copy(string destFolderName)
		{
			if(NMtpCopyFolderFirst(FullPath, destFolderName, CopyAllNoResult) == (IntPtr)INVALID_HANDLE_VALUE) {
				return false;
			}
			return true;
		}

		/// <summary>
		/// フォルダファイルコピー
		/// </summary>
		/// <param name="destFolderName">コピー先フォルダ</param>
		/// <param name="flag">コピー設定</param>
		/// <returns>true で成功</returns>
		public bool Copy(string destFolderName, int flag)
		{
			if(NMtpCopyFolderFirst(FullPath, destFolderName, flag | CopyAllNoResult) == (IntPtr)INVALID_HANDLE_VALUE) {
				return false;
			}
			return true;
		}

		/// <summary>
		/// フォルダファイルコピー
		/// </summary>
		/// <param name="destFolderName">コピー先フォルダ</param>
		/// <param name="flag">コピー設定</param>
		/// <param name="copyPattern">ファイル指定文字列</param>
		/// <returns>true で成功</returns>
		public bool Copy(string destFolderName, int flag, string copyPattern)
		{
			if(NMtpCopyFolderFirst(Path.Combine(FullPath, copyPattern), destFolderName, flag | CopyAllNoResult) == (IntPtr)INVALID_HANDLE_VALUE) {
				return false;
			}
			return true;
		}

		/// <summary>
		/// <para>フォルダファイルコピー</para>
		/// <para>コピーファイルの取得</para>
		/// </summary>
		/// <param name="destFolderName">コピー先フォルダ</param>
		/// <param name="flag">コピー動作設定</param>
		/// <example>
		/// <para>
		/// PC\Xperia Z\内部ストレージ\DCIM\100ANDRO フォルダ内のファイルを C:\Photo にコピー
		/// </para>
		/// <para>
		/// 同じ名前のファイルがある場合は新しいファイルのみコピーする
		/// <code lang="cs">
		/// nMtp.DirectoryInfo dir = new nMtp.DirectoryInfo(@"PC\Xperia Z\内部ストレージ\DCIM\100ANDRO");
		/// try {
		///   dir.CopyFirst(@"C:\Photo", nMtp.DirectoryInfo.CopyNewer);
		///   while(dir.CopyNext()) {
		///      if(dir.CopyResult) {
		///        Console.WriteLine(src + "を" + dst + "にコピーしました");
		///      } else {
		///        Console.WriteLine(src + "を" + dst + "にコピーできませんでした " + dir.CopyErrorString);
		///      }
		///   }
		///   dir.CopyClose();
		/// } catch (IOException ex) {
		///   Constole.WriteLine(ex.Message);
		/// }
		/// </code>
		/// <code lang="vb">
		/// Dim dir As New nMtp.DirectoryInfo("PC\Xperia Z\内部ストレージ\DCIM\100ANDRO")
		/// Try
		/// 	dir.CopyFirst("C:\Photo", nMtp.DirectoryInfo.CopyNewer)
		/// 	While dir.CopyNext()
		/// 		If dir.CopyResult Then
		/// 			Console.WriteLine(src + "を" + dst + "にコピーしました")
		/// 		Else
		/// 			Console.WriteLine(src + "を" + dst + "にコピーできませんでした " + dir.CopyErrorString)
		/// 		End If
		/// 	End While
		/// 	dir.CopyClose()
		/// Catch ex As IOException
		/// 	Constole.WriteLine(ex.Message)
		/// End Try
		/// </code>
		/// </para>
		/// </example>
		/// <exception cref="ArgumentNullException">
		///	フォルダ が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	フォルダ が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	フォルダ がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	フォルダにアクセスできない
		/// </exception>
		public void CopyFirst(string destFolderName, int flag)
		{
			Directory.CheckInvalidPathName(destFolderName, false);

			_handle = NMtpCopyFolderFirst(FullPath, destFolderName, flag);
			if(_handle == (IntPtr)INVALID_HANDLE_VALUE) {
				_handle = IntPtr.Zero;
				int error_code = Marshal.GetLastWin32Error();
				string message = new System.ComponentModel.Win32Exception(error_code).Message;
				if(error_code == ERROR_PATH_NOT_FOUND) {
					throw new DirectoryNotFoundException(message);
				} else {
					throw new IOException(message);
				}
			}
		}

		/// <summary>
		/// <para>フォルダファイルコピー</para>
		/// <para>コピーファイルの取得</para>
		/// </summary>
		/// <param name="destFolderName">コピー先フォルダ</param>
		/// <param name="flag">コピー動作設定</param>
		/// <param name="copyPattern">ファイル指定文字列</param>
		/// <example>
		/// <para>
		/// PC\Xperia Z\内部ストレージ\DCIM\100ANDRO フォルダ内の拡張子が jpg のファイルを C:\Photo にコピー
		/// <code lang="cs">
		/// nMtp.DirectoryInfo dir = new nMtp.DirectoryInfo(@"PC\Xperia Z\内部ストレージ\DCIM\100ANDRO");
		/// try {
		///   dir.CopyFirst(@"C:\Photo", 0, "*.jpg");
		///   while(dir.CopyNext()) {
		///      if(dir.CopyResult) {
		///        Console.WriteLine(dir.CopySourceFile + "を" + dir.CopyDestFile + "にコピーしました");
		///      } else {
		///        Console.WriteLine(dir.CopySourceFile + "を" + dir.CopyDestFile + "にコピーできませんでした " + dir.CopyErrorString);
		///      }
		///   }
		///   dir.CopyClose();
		/// } catch (IOException ex) {
		///   Constole.WriteLine(ex.Message);
		/// }
		/// </code>
		/// <code lang="vb">
		/// Dim dir As New nMtp.DirectoryInfo("PC\Xperia Z\内部ストレージ\DCIM\100ANDRO")
		/// Try
		/// 	dir.CopyFirst("C:\Photo", 0, "*.jpg")
		/// 	While dir.CopyNext()
		/// 		If dir.CopyResult Then
		/// 			Console.WriteLine(dir.CopySourceFile + "を" + dir.CopyDestFile + "にコピーしました")
		/// 		Else
		/// 			Console.WriteLine(dir.CopySourceFile + "を" + dir.CopyDestFile + "にコピーできませんでした " + dir.CopyErrorString)
		/// 		End If
		/// 	End While
		/// 	dir.CopyClose()
		/// Catch ex As IOException
		/// 	Constole.WriteLine(ex.Message)
		/// End Try
		/// </code>
		/// </para>
		/// </example>
		/// <exception cref="ArgumentNullException">
		///	フォルダ が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	フォルダ が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	フォルダ がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	フォルダにアクセスできない
		/// </exception>
		public void CopyFirst(string destFolderName, int flag, string copyPattern)
		{
			Directory.CheckInvalidPathName(destFolderName, false);

			_handle = NMtpCopyFolderFirst(Path.Combine(FullPath, copyPattern), destFolderName, flag);
			if(_handle == (IntPtr)INVALID_HANDLE_VALUE) {
				_handle = IntPtr.Zero;
				int error_code = Marshal.GetLastWin32Error();
				string message = new System.ComponentModel.Win32Exception(error_code).Message;
				if(error_code == ERROR_PATH_NOT_FOUND) {
					throw new DirectoryNotFoundException(message);
				} else {
					throw new IOException(message);
				}
			}
		}

		/// <summary>
		/// <para>フォルダファイルコピー</para>
		/// <para>コピー実行</para>
		/// </summary>
		/// <returns>true でコピーファイル有</returns>
		public bool CopyNext()
		{
			_source_file = "";
			_dest_file = "";
			_copy_result = 0;
			_copy_error = 0;
			if(_handle != IntPtr.Zero) {
				StringBuilder src = new StringBuilder(MAX_PATH);
				StringBuilder dst = new StringBuilder(MAX_PATH);
				if(NMtpCopyFolderNext(_handle, src, MAX_PATH, dst, MAX_PATH, out _copy_result, out _copy_error)) {
					_source_file = src.ToString();
					_dest_file = dst.ToString();
					return true;
				}
			}
			return false;
		}

		/// <summary>
		/// <para>フォルダファイルコピー</para>
		/// <para>終了</para>
		/// </summary>
		public void CopyClose()
		{
			if(_handle != IntPtr.Zero) {
				NMtpCopyFolderClose(_handle);
				_handle = IntPtr.Zero;
			}
		}

		/// <summary>
		/// <para>フォルダファイルコピー</para>
		/// <para>エラー番号取得</para>
		/// </summary>
		public int CopyError
		{
			get {
				return _copy_error;
			}
		}

		/// <summary>
		/// <para>フォルダファイルコピー</para>
		/// <para>コピー元ファイル取得</para>
		/// </summary>
		public string CopySourceFile
		{
			get {
				return _source_file;
			}
		}

		/// <summary>
		/// <para>フォルダファイルコピー</para>
		/// <para>コピー結果</para>
		/// </summary>
		public bool CopyResult
		{
			get {
				if(_copy_result != 0) {
					return true;
				}
				return false;
			}
		}

		/// <summary>
		/// <para>フォルダファイルコピー</para>
		/// <para>コピー先ファイル取得</para>
		/// </summary>
		public string CopyDestFile
		{
			get {
				return _dest_file;
			}
		}
		/// <summary>
		/// <para>フォルダファイルコピー</para>
		/// <para>エラー文字列取得</para>
		/// </summary>
		public string CopyErrorString
		{
			get {
				return Directory.GetCopyErrorString(_copy_error);
			}
		}

		/// <summary>
		/// フォルダファイルコピーのファイル数
		/// </summary>
		public int CopyCount
		{
			get
			{
				if(_handle == IntPtr.Zero)
				{
					return 0;
				}
				else
				{
					return NMtpCopyFolderCount(_handle);
				}
			}
		}

		/// <summary>
		/// 親フォルダ名の取得
		/// </summary>
		public string Parent {
			get {
				return Path.GetDirectoryName(FullPath);
			}
		}
		/// <summary>
		/// ルートパスの取得
		/// </summary>
		public string Root {
			get {
				return Path.GetPathRoot(FullPath);
			}
		}

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpFindFirstFile", CharSet=CharSet.Unicode)]
		protected static extern IntPtr NMtpFindFirstFile(string file_name, out WIN32_FIND_DATA find_data);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpFindNextFile", CharSet=CharSet.Unicode)]
		protected static extern bool NMtpFindNextFile(IntPtr handle, out WIN32_FIND_DATA find_data);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpFindClose")]
		protected static extern bool NMtpFindClose(IntPtr handle);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpCopyFolderFirst", CharSet=CharSet.Unicode, SetLastError=true)]
		protected static extern IntPtr NMtpCopyFolderFirst(string src, string dst, int flag);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpCopyFolderCount")]
		protected static extern int NMtpCopyFolderCount(IntPtr handle);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpCopyFolderNext", CharSet=CharSet.Unicode)]
		protected static extern bool NMtpCopyFolderNext(IntPtr handle, StringBuilder src_name, int src_length, StringBuilder dst_name, int dst_length, out int result, out int error);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpCopyFolderClose")]
		protected static extern int NMtpCopyFolderClose(IntPtr handle);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpWildcardMatch", CharSet=CharSet.Unicode)]
		protected static extern bool NMtpWildcardMatch(string pattern, string name);
	}

	/// <summary>
	/// ファイル操作クラス
	/// </summary>
	[SerializableAttribute]
	[ComVisibleAttribute(true)]
	public class FileInfo : FileSystemInfo {
		/// <summary>
		/// <para>コピー先にファイルがあれば失敗</para>
		/// <para>CopyTo の flag で指定</para>
		/// </summary>
		public const int CopyFailExists = 0x00000001;
		/// <summary>
		/// <para>新しいファイルのみコピー</para>
		/// <para>CopyTo の flag で指定</para>
		/// </summary>
		public const int CopyNewer = 0x00000010;

		private const int INVALID_HANDLE_VALUE = -1;
		private const int MAX_DIR_PATH = 248;

		/// <summary>
		/// <c>FileInfo</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		/// <param name="path">ファイル名</param>
		public FileInfo(string path) : base(path) {
		}

		/// <summary>
		/// <c>FileInfo</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		/// <param name="path">ファイル名</param>
		/// <param name="find_data">ファイルの情報</param>
		public FileInfo(string path, WIN32_FIND_DATA find_data) : base(path, find_data) {
		}

		/// <summary>
		/// ファイルの削除
		/// </summary>
		///	<exception cref="IOException">
		///	ファイルを削除できない
		/// </exception>
		public override void Delete() {
			File.Delete(FullPath);
		}

		/// <summary>
		/// ファイルのコピー
		/// </summary>
		/// <param name="destFileName">コピー先のファイルパス</param>
		///	<exception cref="IOException">
		///	ファイルをコピーできない
		/// </exception>
		public void CopyTo(string destFileName)
		{
			CopyTo(destFileName, false);
		}

		/// <summary>
		/// ファイルのコピー
		/// </summary>
		/// <param name="destFileName">コピー先のファイルパス</param>
		/// <param name="overwrite">true で上書</param>
		///	<exception cref="IOException">
		///	ファイルをコピーできない
		/// </exception>
		public void CopyTo(string destFileName, bool overwrite)
		{
			File.Copy(FullPath, destFileName, overwrite);
		}

		/// <summary>
		/// ファイルのコピー
		/// </summary>
		/// <param name="destFileName">コピー先のファイルパス</param>
		/// <param name="flag">コピー設定</param>
		///	<exception cref="IOException">
		///	ファイルをコピーできない
		/// </exception>
		public void CopyTo(string destFileName, int flag)
		{
			File.Copy(FullPath, destFileName, flag);
		}

		/// <summary>
		/// ファイルの移動
		/// </summary>
		/// <param name="destFileName">移動先のファイルパス</param>
		///	<exception cref="IOException">
		///	ファイルを移動できない
		/// </exception>
		public void MoveTo(string destFileName)
		{
			File.Move(FullPath, destFileName);
		}

		/// <summary>
		/// ファイルが読出し専用かどうかを示す値を取得
		/// </summary>
		public bool IsReadOnly {
			get {
				GetStatus();
				if((_attributes & FileAttributes.ReadOnly) == FileAttributes.ReadOnly) {
					return true;
				}
				return false;
			}
		}


		/// <summary>
		/// 親フォルダのインスタンスを取得
		/// </summary>
		public DirectoryInfo Directory {
			get {
				return new DirectoryInfo(Path.GetDirectoryName(FullPath));
			}
		}

		/// <summary>
		/// 親フォルダ名を取得
		/// </summary>
		public string DirectoryName {
			get {
				return Path.GetDirectoryName(FullPath);
			}
		}

		/// <summary>
		/// ファイルサイズを取得
		/// </summary>
		public long Length {
			get {
				GetStatus();
				return _length;
			}
		}

	}

	/// <summary>
	/// フォルダ操作クラス
	/// </summary>
	public class Directory
	{
		/// <summary>
		/// <para>コピー先にファイルがあれば失敗</para>
		/// <para>Copy/CopyFirst の flag で指定</para>
		/// </summary>
		public const int CopyFailExists = 0x00000001;
		/// <summary>
		/// <para>新しいファイルのみコピー</para>
		/// <para>Copy/CopyFirst の flag で指定</para>
		/// </summary>
		public const int CopyNewer = 0x00000010;
		/// <summary>
		/// <para>サブフォルダの中身もコピー</para>
		/// <para>Copy/CopyFirst の flag で指定</para>
		/// </summary>
		public const int CopySubFolder = 0x00000100;
		/// <summary>
		/// <para>一度にコピー</para>
		/// <para>CopyFirst の flag で指定</para>
		/// </summary>
		public const int CopyAll = 0x00001000;
		/// <summary>
		/// <para>一度にコピーして結果も保存しない</para>
		/// </summary>
		public const int CopyAllNoResult = 0x00002000;

		private const int ERROR_FILE_NOT_FOUND = 2;
		private const int ERROR_PATH_NOT_FOUND = 3;
		private const int ERROR_ACCESS_DENIED = 5;
		private const int ERROR_FILE_EXISTS = 80;
		private const int ERROR_CANNOT_MAKE = 82;

		private const int INVALID_HANDLE_VALUE = -1;
		private const int MAX_OBJECT_ID = 14;
		private const int MAX_PATH = 260;
		private const int MAX_DIR_PATH = 248;

		private IntPtr _handle = IntPtr.Zero;

		private const uint FILE_ATTRIBUTE_DIRECTORY = 0x00000010;

		/// <summary>
		/// <c>Directory</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		public Directory() {
		}

		/// <summary>
		/// MTP 接続デバイス一覧取得
		/// </summary>
		/// <returns>デバイス名一覧</returns>
		public static string [] GetDevices()
		{
			IntPtr h;
			StringBuilder name;
			name = new StringBuilder(MAX_PATH);
			List<string> list = new List<string>();

			h = NMtpFindDeviceFirst();
			if(h != (IntPtr)INVALID_HANDLE_VALUE) {
				while(NMtpFindDeviceNext(h, name, MAX_PATH)) {
					list.Add(name.ToString());
				}
			}
			return list.ToArray();
		}

		/// <summary>
		/// パス名のチェック
		/// </summary>
		/// <param name="path">パス名</param>
		/// <param name="wildcard">ファイル名にワイルドカードが入る</param>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		public static void CheckInvalidPathName(string path, bool wildcard)
		{
			if(path == null) {
				throw new ArgumentNullException();
			}
			if(path.Length >= MAX_DIR_PATH) {
				throw new PathTooLongException();
			}
			string space = path.Trim();
			if(space == String.Empty) {
				throw new ArgumentNullException();
			}
			string invalid_chars = "\"<>|/\0\x0001\x0002\x0003\x0004\x0005\x0006\x0007\x0008\x0009\x000a\x000b\x000c\x000d\x000e\x000f\x0010\x0011\x0012\x0013\x0014\x0015\x0016\x0017\x0018\x0019\x001a\x001b\x001c\x001d\x001e\x001f";
			if(path.IndexOfAny(invalid_chars.ToCharArray()) >= 0) {
				throw new ArgumentException();
			}
			string dir = Path.GetDirectoryName(path);
			invalid_chars = "*?";
			if(dir != null) {
				if(dir.IndexOfAny(invalid_chars.ToCharArray()) >= 0) {
					throw new ArgumentException();
				}
			}
			string name = Path.GetFileName(path);
			if(name != null) {
				if(wildcard) {
					invalid_chars = ":\\";
				} else {
					invalid_chars += ":\\";
				}
				if(name.IndexOfAny(invalid_chars.ToCharArray()) >= 0) {
					throw new ArgumentException();
				}
			}
		}

		/// <summary>
		/// 指定されたフォルダ内のサブフォルダの名前を取得
		/// </summary>
		/// <param name="path">検索するフォルダのパス</param>
		/// <returns>指定したパス内のサブフォルダ名の配列。フォルダが見つからない場合は空の配列</returns>
		public static string [] GetDirectories(string path)
		{
			return GetDirectories(path, "*");
		}

		/// <summary>
		/// 指定されたフォルダ内のサブフォルダの名前を取得
		/// </summary>
		/// <param name="path">検索するフォルダのパス</param>
		/// <param name="searchPattern">検索文字列</param>
		/// <returns>指定したパス内のサブフォルダ名の配列。フォルダが見つからない場合は空の配列</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		public static string [] GetDirectories(string path, string searchPattern)
		{
			CheckInvalidPathName(path, false);

			IntPtr h;
			WIN32_FIND_DATA find_data;
			List<string> list = new List<string>();

			h = NMtpFindFirstFile(Path.Combine(path, searchPattern), out find_data);
			if(h != (IntPtr)INVALID_HANDLE_VALUE) {
				do {
					if((find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY) {
						list.Add(Path.Combine(path, find_data.cFileName.ToString()));
					}
				} while(NMtpFindNextFile(h, out find_data));
				NMtpFindClose(h);
			}
			return list.ToArray();
		}

		/// <summary>
		/// 指定されたフォルダ内のファイル名を取得
		/// </summary>
		/// <param name="path">検索するフォルダのパス</param>
		/// <returns>指定したパス内のサブフォルダのファイル名の配列。ファイルが見つからない場合は空の配列</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		public static string [] GetFiles(string path)
		{
			return GetFiles(path, "*");
		}

		/// <summary>
		/// 指定されたフォルダ内のファイル名を取得
		/// </summary>
		/// <param name="path">検索するフォルダのパス</param>
		/// <param name="searchPattern">検索文字列</param>
		/// <returns>指定したパス内のサブフォルダのファイル名の配列。ファイルが見つからない場合は空の配列</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		public static string [] GetFiles(string path, string searchPattern)
		{
			CheckInvalidPathName(path, false);

			IntPtr h;
			WIN32_FIND_DATA find_data;
			List<string> list = new List<string>();

			h = NMtpFindFirstFile(Path.Combine(path, searchPattern), out find_data);
			if(h != (IntPtr)INVALID_HANDLE_VALUE) {
				do {
					if((find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0) {
						list.Add(Path.Combine(path, find_data.cFileName.ToString()));
					}
				} while(NMtpFindNextFile(h, out find_data));
				NMtpFindClose(h);
			}
			return list.ToArray();
		}

		/// <summary>
		/// 指定されたフォルダ内のファイルとフォルダの名前の取得
		/// </summary>
		/// <param name="path">検索するフォルダのパス</param>
		/// <returns>指定したパス内のファイルとサブフォルダの名前の配列。ファイルが見つからない場合は空の配列</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		public static string [] GetFileSystemEntries(string path)
		{
			return GetFileSystemEntries(path, "*");
		}

		/// <summary>
		/// 指定されたフォルダ内のファイルとフォルダの名前の取得
		/// </summary>
		/// <param name="path">検索するフォルダのパス</param>
		/// <param name="searchPattern">検索文字列</param>
		/// <returns>指定したパス内のファイルとサブフォルダの名前の配列。ファイルが見つからない場合は空の配列</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		public static string [] GetFileSystemEntries(string path, string searchPattern)
		{
			CheckInvalidPathName(path, false);

			IntPtr h;
			WIN32_FIND_DATA find_data;
			List<string> list = new List<string>();

			h = NMtpFindFirstFile(Path.Combine(path, searchPattern), out find_data);
			if(h != (IntPtr)INVALID_HANDLE_VALUE) {
				do {
					list.Add(Path.Combine(path, find_data.cFileName.ToString()));
				} while(NMtpFindNextFile(h, out find_data));
				NMtpFindClose(h);
			}
			return list.ToArray();
		}

		/// <summary>
		/// フォルダの作成
		/// </summary>
		/// <param name="path">作成するフォルダ</param>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	フォルダが作成できない
		/// </exception>
		public static void CreateDirectory(string path)
		{
			CheckInvalidPathName(path, false);

			if(!NMtpCreateFolder(path)) {
				int error_code = Marshal.GetLastWin32Error();
				string message = new System.ComponentModel.Win32Exception(error_code).Message;
				if(error_code == ERROR_PATH_NOT_FOUND) {
					throw new DirectoryNotFoundException(message);
				} else {
					throw new IOException(message);
				}
			}
		}

		/// <summary>
		/// フォルダの削除
		/// </summary>
		/// <param name="path">削除するフォルダ</param>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	フォルダが作成できない
		/// </exception>
		public static void Delete(string path)
		{
			CheckInvalidPathName(path, false);

			if(!NMtpRemoveFolder(path)) {
				int error_code = Marshal.GetLastWin32Error();
				string message = new System.ComponentModel.Win32Exception(error_code).Message;
				if(error_code == ERROR_PATH_NOT_FOUND) {
					throw new DirectoryNotFoundException(message);
				} else {
					throw new IOException(message);
				}
			}
		}

		/// <summary>
		/// フォルダの削除
		/// </summary>
		/// <param name="path">削除するフォルダ</param>
		/// <param name="recursive">true で指定されたフォルダのファイル、サブフォルダも削除</param>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	フォルダが作成できない
		/// </exception>
		public static void Delete(string path, bool recursive)
		{
			CheckInvalidPathName(path, false);

			if(recursive) {
				if(!NMtpDeleteFolder(path)) {
					int error_code = Marshal.GetLastWin32Error();
					string message = new System.ComponentModel.Win32Exception(error_code).Message;
					if(error_code == ERROR_PATH_NOT_FOUND) {
						throw new DirectoryNotFoundException(message);
					} else {
						throw new IOException(message);
					}
				}
			} else {
				if(!NMtpRemoveFolder(path)) {
					int error_code = Marshal.GetLastWin32Error();
					string message = new System.ComponentModel.Win32Exception(error_code).Message;
					if(error_code == ERROR_PATH_NOT_FOUND) {
						throw new DirectoryNotFoundException(message);
					} else {
						throw new IOException(message);
					}
				}
			}
		}

		/// <summary>
		/// フォルダの存在確認
		/// </summary>
		/// <param name="path">フォルダ</param>
		/// <returns>true で存在</returns>
		public static bool Exists(string path)
		{
			CheckInvalidPathName(path, false);

			WIN32_FIND_DATA find_data;

			if(NMtpGetStatus(path, out find_data)) {
				if((find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY) {
					return true;
				}
			}
			return false;
		}

		private const int timeCreation = 0;
		private const int timeLastWrite = 1;
		private const int timeLastAccess = 2;

		private static DateTime GetTime(string path, int time_type, bool local_flag)
		{
			CheckInvalidPathName(path, false);

			WIN32_FIND_DATA find_data;
			DateTime datetime = new DateTime();

			if(NMtpGetStatus(path, out find_data)) {
				if((find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY) {
					if(time_type == timeLastWrite) {
						datetime = DateTime.FromFileTime((((long)find_data.ftLastWriteTime.dwHighDateTime) << 32) + find_data.ftLastWriteTime.dwLowDateTime);
					} else if(time_type == timeCreation) {
						datetime = DateTime.FromFileTime((((long)find_data.ftCreationTime.dwHighDateTime) << 32) + find_data.ftCreationTime.dwLowDateTime);
					} else if(time_type == timeLastAccess) {
						datetime = DateTime.FromFileTime((((long)find_data.ftLastAccessTime.dwHighDateTime) << 32) + find_data.ftLastAccessTime.dwLowDateTime);
					}
				} else {
					string message = new System.ComponentModel.Win32Exception(ERROR_PATH_NOT_FOUND).Message;
					throw new DirectoryNotFoundException(message);
				}
			} else {
				int error_code = Marshal.GetLastWin32Error();
				string message = new System.ComponentModel.Win32Exception(error_code).Message;
				if(error_code == ERROR_PATH_NOT_FOUND) {
					throw new DirectoryNotFoundException(message);
				} else {
					throw new IOException(message);
				}
			}
			if(local_flag) {
				datetime.ToLocalTime();
			}
			return datetime;
		}

		/// <summary>
		/// フォルダの作成日時を取得
		/// </summary>
		/// <param name="path">フォルダ</param>
		/// <returns>作成日時</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	フォルダにアクセスできない
		/// </exception>
		public static DateTime GetCreationTime(string path)
		{
			return GetTime(path, timeCreation, true);
		}

		/// <summary>
		/// フォルダの作成日時(世界協定時刻)を取得
		/// </summary>
		/// <param name="path">フォルダ</param>
		/// <returns>作成日時(世界協定時刻)</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	フォルダにアクセスできない
		/// </exception>
		public static DateTime GetCreationTimeUtc(string path)
		{
			return GetTime(path, timeCreation, false);
		}

		/// <summary>
		/// フォルダの最後の書込日時を取得
		/// </summary>
		/// <param name="path">フォルダ</param>
		/// <returns>書込日時</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	フォルダにアクセスできない
		/// </exception>
		public static DateTime GetLastWriteTime(string path)
		{
			return GetTime(path, timeLastWrite, true);
		}

		/// <summary>
		/// フォルダの最後の書込日時(世界協定時刻)を取得
		/// </summary>
		/// <param name="path">フォルダ</param>
		/// <returns>書込日時(世界協定時刻)</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	フォルダにアクセスできない
		/// </exception>
		public static DateTime GetLastWriteTimeUtc(string path)
		{
			return GetTime(path, timeLastWrite, false);
		}

		/// <summary>
		/// フォルダの最後のアクセス日時を取得
		/// </summary>
		/// <param name="path">フォルダ</param>
		/// <returns>アクセス日時</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	フォルダにアクセスできない
		/// </exception>
		public static DateTime GetLastAccessTime(string path)
		{
			return GetTime(path, timeLastAccess, true);
		}

		/// <summary>
		/// フォルダの最後のアクセス日時(世界協定時刻)を取得
		/// </summary>
		/// <param name="path">フォルダ</param>
		/// <returns>アクセス日時(世界協定時刻)</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	フォルダにアクセスできない
		/// </exception>
		public static DateTime GetLastAccessTimeUtc(string path)
		{
			return GetTime(path, timeLastAccess, false);
		}

		/// <summary>
		/// フォルダファイルコピー
		/// </summary>
		/// <param name="sourceFolderName">コピー元フォルダ</param>
		/// <param name="destFolderName">コピー先フォルダ</param>
		/// <returns>true で成功</returns>
		/// <example>
		/// <para>
		/// PC\Xperia Z\内部ストレージ\DCIM\100ANDRO フォルダ内の拡張子が jpg のファイルを C:\Photo にコピー
		/// <code lang="cs">
		///   nMtp.Directory.Copy(@"PC\Xperia Z\内部ストレージ\DCIM\100ANDRO\*.jpg", @"C:\Photo");
		/// </code>
		/// <code lang="vb">
		///  nMtp.Directory.Copy("PC\Xperia Z\内部ストレージ\DCIM\100ANDRO\*.jpg", "C:\Photo")
		/// </code>
		/// </para>
		/// </example>
		public static bool Copy(string sourceFolderName, string destFolderName)
		{
			return Copy(sourceFolderName, destFolderName, 0);
		}

		/// <summary>
		/// フォルダファイルコピー
		/// </summary>
		/// <param name="sourceFolderName">コピー元フォルダ</param>
		/// <param name="destFolderName">コピー先フォルダ</param>
		/// <param name="flag">コピー設定</param>
		/// <returns>true で成功</returns>
		/// <example>
		/// <para>
		/// PC\Xperia Z\内部ストレージ\DCIM\100ANDRO フォルダ内のファイルを C:\Photo にコピー
		/// </para>
		/// <para>
		/// 同じ名前のファイルがある場合は新しいファイルのみコピーする
		/// <code lang="cs">
		///   nMtp.Directory.Copy(@"PC\Xperia Z\内部ストレージ\DCIM\100ANDRO", @"C:\Photo", nMtp.Directory.CopyNewer);
		/// </code>
		/// <code lang="vb">
		///   nMtp.Directory.Copy("PC\Xperia Z\内部ストレージ\DCIM\100ANDRO", "C:\Photo", nMtp.Directory.CopyNewer)
		/// </code>
		/// </para>
		/// </example>
		public static bool Copy(string sourceFolderName, string destFolderName, int flag)
		{
			CheckInvalidPathName(sourceFolderName, true);
			CheckInvalidPathName(destFolderName, false);

			if(NMtpCopyFolderFirst(sourceFolderName, destFolderName, flag | CopyAllNoResult) == (IntPtr)INVALID_HANDLE_VALUE) {
				return false;
			}
			return true;
		}

		/// <summary>
		/// <para>フォルダファイルコピー</para>
		/// <para>コピーファイルの取得</para>
		/// </summary>
		/// <param name="sourceFolderName">コピー元フォルダ</param>
		/// <param name="destFolderName">コピー先フォルダ</param>
		/// <param name="flag">コピー動作設定</param>
		/// <example>
		/// <para>
		/// PC\Xperia Z\内部ストレージ\DCIM フォルダ内の中身をサブフォルダも含めて C:\Photo にコピー
		/// <code lang="cs">
		/// nMtp.Directory dir = new nMtp.Directory();
		/// try {
		///   dir.CopyFirst(@"PC\Xperia Z\内部ストレージ\DCIM", @"C:\Photo", nMtp.Directory.CopySubFolder);
		///   while(dir.CopyNext(out src, out dst, out result, out error)) {
		///      if(result != 0) {
		///        Console.WriteLine(src + "を" + dst + "にコピーしました");
		///      } else {
		///        Console.WriteLine(src + "を" + dst + "にコピーできませんでした " + nMtp.Directory.GetCopyErrorString(error));
		///      }
		///   }
		///   dir.CopyClose();
		/// } catch (IOException ex) {
		///   Constole.WriteLine(ex.Message);
		/// }
		/// </code>
		/// <code lang="vb">
		/// Dim dir As New nMtp.Directory()
		/// Try
		///   dir.CopyFirst("PC\Xperia Z\内部ストレージ\DCIM", "C:\Photo", nMtp.Directory.CopySubFolder)
		///   While dir.CopyNext(src, dst, result, error_code)
		///     If result &lt;&gt; 0 Then
		///       Console.WriteLine(src + "を" + dst + "にコピーしました")
		///     Else
		///       Console.WriteLine(src + "を" + dst + "にコピーできませんでした " + nMtp.Directory.GetCopyErrorString(error_code))
		///     End If
		///   End While
		///   dir.CopyClose()
		/// Catch ex As IOException
		///   Constole.WriteLine(ex.Message)
		/// End Try
		/// </code>
		/// </para>
		/// </example>
		/// <exception cref="ArgumentNullException">
		///	フォルダ が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	フォルダ が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	フォルダ がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	フォルダにアクセスできない
		/// </exception>
		public void CopyFirst(string sourceFolderName, string destFolderName, int flag)
		{
			CheckInvalidPathName(sourceFolderName, true);
			CheckInvalidPathName(destFolderName, false);

			_handle = NMtpCopyFolderFirst(sourceFolderName, destFolderName, flag);
			if(_handle == (IntPtr)INVALID_HANDLE_VALUE) {
				_handle = IntPtr.Zero;
				int error_code = Marshal.GetLastWin32Error();
				string message = new System.ComponentModel.Win32Exception(error_code).Message;
				if(error_code == ERROR_PATH_NOT_FOUND) {
					throw new DirectoryNotFoundException(message);
				} else {
					throw new IOException(message);
				}
			}
		}

		/// <summary>
		/// <para>フォルダファイルコピー</para>
		/// <para>コピー実行</para>
		/// </summary>
		/// <param name="sourceFileName">コピー元のファイル</param>
		/// <param name="destFileName">コピー先のファイル</param>
		/// <param name="result">コピー結果</param>
		/// <param name="error">コピー失敗時のエラーコード</param>
		/// <returns>true でコピーファイル有</returns>
		public bool CopyNext(out string sourceFileName, out string destFileName, out int result, out int error)
		{
			sourceFileName = "";
			destFileName = "";
			result = 0;
			error = 0;
			if(_handle != IntPtr.Zero) {
				StringBuilder src = new StringBuilder(MAX_PATH);
				StringBuilder dst = new StringBuilder(MAX_PATH);
				if(NMtpCopyFolderNext(_handle, src, MAX_PATH, dst, MAX_PATH, out result, out error)) {
					sourceFileName = src.ToString();
					destFileName = dst.ToString();
					return true;
				}
			}
			return false;
		}

		/// <summary>
		/// <para>フォルダファイルコピー</para>
		/// <para>終了</para>
		/// </summary>
		public void CopyClose()
		{
			if(_handle != IntPtr.Zero) {
				NMtpCopyFolderClose(_handle);
				_handle = IntPtr.Zero;
			}
		}

		/// <summary>
		/// <para>フォルダファイルコピー</para>
		/// <para>エラー文字列取得</para>
		/// </summary>
		/// <param name="error_code">エラーコード</param>
		/// <returns>エラー文字列</returns>
		public static string GetCopyErrorString(int error_code)
		{
			return new System.ComponentModel.Win32Exception(error_code).Message;
		}

		/// <summary>
		/// フォルダファイルコピーのファイル数
		/// </summary>
		public int CopyCount
		{
			get
			{
				if(_handle == IntPtr.Zero)
				{
					return 0;
				}
				else
				{
					return NMtpCopyFolderCount(_handle);
				}
			}
		}

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpCopyFile", CharSet=CharSet.Unicode, SetLastError=true)]
		protected static extern bool NMtpCopyFile(string src, string dst, int fail_exists);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpMoveFile", CharSet=CharSet.Unicode, SetLastError=true)]
		protected static extern bool NMtpMoveFile(string src, string dst);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpDeleteFile", CharSet=CharSet.Unicode, SetLastError=true)]
		protected static extern bool NMtpDeleteFile(string path);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpCreateFolder", CharSet=CharSet.Unicode, SetLastError=true)]
		protected static extern bool NMtpCreateFolder(string path);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpDeleteFolder", CharSet=CharSet.Unicode, SetLastError=true)]
		protected static extern bool NMtpDeleteFolder(string path);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpRemoveFolder", CharSet=CharSet.Unicode, SetLastError=true)]
		protected static extern bool NMtpRemoveFolder(string path);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpFindFirstFile", CharSet=CharSet.Unicode)]
		protected static extern IntPtr NMtpFindFirstFile(string file_name, out WIN32_FIND_DATA find_data);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpFindNextFile", CharSet=CharSet.Unicode)]
		protected static extern bool NMtpFindNextFile(IntPtr handle, out WIN32_FIND_DATA find_data);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpFindClose")]
		protected static extern bool NMtpFindClose(IntPtr handle);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpFindDeviceFirst")]
		protected static extern IntPtr NMtpFindDeviceFirst();

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpFindDeviceNext", CharSet=CharSet.Unicode)]
		protected static extern bool NMtpFindDeviceNext(IntPtr handle, StringBuilder name, int length);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpFindDeviceClose")]
		protected static extern bool NMtpFindDeviceClose(IntPtr handle);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpGetStatus", CharSet=CharSet.Unicode, SetLastError=true)]
		protected static extern bool NMtpGetStatus(string path, out WIN32_FIND_DATA find_data);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpCopyFolderFirst", CharSet=CharSet.Unicode, SetLastError=true)]
		protected static extern IntPtr NMtpCopyFolderFirst(string src, string dst, int flag);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpCopyFolderCount")]
		protected static extern int NMtpCopyFolderCount(IntPtr handle);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpCopyFolderNext", CharSet=CharSet.Unicode)]
		protected static extern bool NMtpCopyFolderNext(IntPtr handle, StringBuilder src_name, int src_length, StringBuilder dst_name, int dst_length, out int result, out int error);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpCopyFolderClose")]
		protected static extern int NMtpCopyFolderClose(IntPtr handle);
	}

	/// <summary>
	/// ファイル操作クラス
	/// </summary>
	public class File
	{
		/// <summary>
		/// <para>コピー先にファイルがあれば失敗</para>
		/// <para>Copy の flag で指定</para>
		/// </summary>
		public const int CopyFailExists = 0x00000001;
		/// <summary>
		/// <para>新しいファイルのみコピー</para>
		/// <para>Copy の flag で指定</para>
		/// </summary>
		public const int CopyNewer = 0x00000010;

		private const int ERROR_FILE_NOT_FOUND = 2;
		private const int ERROR_PATH_NOT_FOUND = 3;
		private const int ERROR_ACCESS_DENIED = 5;
		private const int ERROR_FILE_EXISTS = 80;
		private const int ERROR_CANNOT_MAKE = 82;

		private const int MAX_PATH = 260;

		private const uint FILE_ATTRIBUTE_DIRECTORY = 0x00000010;

		/// <summary>
		/// <c>File</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		public File() {
		}

		/// <summary>
		/// ファイルコピー
		/// </summary>
		/// <param name="sourceFileName">コピー元ファイル</param>
		/// <param name="destFileName">コピー先ファイル</param>
		/// <exception cref="ArgumentNullException">
		///	ファイルパス が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	ファイルパス が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	ファイルパス がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="FileNotFoundException">
		///	指定されたファイルが存在しない
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	ファイルにアクセスできない
		/// </exception>
		public static void Copy(string sourceFileName, string destFileName)
		{
			Copy(sourceFileName, destFileName, true);
		}

		/// <summary>
		/// ファイルコピー
		/// </summary>
		/// <param name="sourceFileName">コピー元ファイル</param>
		/// <param name="destFileName">コピー先ファイル</param>
		/// <param name="overwrite">true で上書き</param>
		/// <exception cref="ArgumentNullException">
		///	ファイルパス が null
		///	</exception>
		///	<exception cref="ArgumentException">
		/// ファイルパス が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	ファイルパス がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="FileNotFoundException">
		///	指定されたファイルが存在しない
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	ファイルにアクセスできない
		/// </exception>
		public static void Copy(string sourceFileName, string destFileName, bool overwrite)
		{
			Copy(sourceFileName, destFileName, (overwrite) ? 0 : CopyFailExists);
		}

		/// <summary>
		/// ファイルコピー
		/// </summary>
		/// <param name="sourceFileName">コピー元ファイル</param>
		/// <param name="destFileName">コピー先ファイル</param>
		/// <param name="flag">コピー動作設定</param>
		/// <example>
		/// <para>
		/// PC\Xperia Z\内部ストレージ\DCIM\100ANDRO\DSC_0001.jpg を C:\Photo\DSC_0001.jpg にコピー
		/// </para>
		/// <para>
		/// 同じ名前のファイルがある場合、新しければコピーする
		/// <code lang="cs">
		///   nMtp.File.Copy(@"PC\Xperia Z\内部ストレージ\DCIM\100ANDRO\DSC_0001.jpg", @"C:\Photo\DSC_0001.jpg", nMtp.File.CopyNewer);
		/// </code>
		/// <code lang="vb">
		///   nMtp.File.Copy("PC\Xperia Z\内部ストレージ\DCIM\100ANDRO\DSC_0001.jpg", "C:\Photo\DSC_0001.jpg", nMtp.File.CopyNewer)
		/// </code>
		/// </para>
		/// </example>
		/// <exception cref="ArgumentNullException">
		///	ファイルパス が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	ファイルパス が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	ファイルパス がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="FileNotFoundException">
		///	指定されたファイルが存在しない
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	ファイルにアクセスできない
		/// </exception>
		public static void Copy(string sourceFileName, string destFileName, int flag)
		{
			Directory.CheckInvalidPathName(sourceFileName, false);
			Directory.CheckInvalidPathName(destFileName, false);

			if(!NMtpCopyFile(sourceFileName, destFileName, flag)) {
				int error_code = Marshal.GetLastWin32Error();
				string message = new System.ComponentModel.Win32Exception(error_code).Message;
				if(error_code == ERROR_FILE_NOT_FOUND) {
					throw new FileNotFoundException(message);
				} else if(error_code == ERROR_PATH_NOT_FOUND) {
					throw new DirectoryNotFoundException(message);
				} else {
					throw new IOException(message);
				}
			}
		}

		/// <summary>
		/// ファイル移動
		/// </summary>
		/// <param name="sourceFileName">移動元ファイル</param>
		/// <param name="destFileName">移動先ファイル</param>
		/// <exception cref="ArgumentNullException">
		///	ファイルパス が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	ファイルパス が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	ファイルパス がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="FileNotFoundException">
		///	指定されたファイルが存在しない
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	ファイルにアクセスできない
		/// </exception>
		public static void Move(string sourceFileName, string destFileName)
		{
			Directory.CheckInvalidPathName(sourceFileName, false);
			Directory.CheckInvalidPathName(destFileName, false);

			if(!NMtpMoveFile(sourceFileName, destFileName)) {
				int error_code = Marshal.GetLastWin32Error();
				string message = new System.ComponentModel.Win32Exception(error_code).Message;
				if(error_code == ERROR_FILE_NOT_FOUND) {
					throw new FileNotFoundException(message);
				} else if(error_code == ERROR_PATH_NOT_FOUND) {
					throw new DirectoryNotFoundException(message);
				} else {
					throw new IOException(message);
				}
			}
		}

		/// <summary>
		/// ファイル削除
		/// </summary>
		/// <param name="path">削除するファイル</param>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="FileNotFoundException">
		///	指定されたファイルが存在しない
		/// </exception>
		///	<exception cref="IOException">
		///	ファイルにアクセスできない
		/// </exception>
		public static void Delete(string path)
		{
			Directory.CheckInvalidPathName(path, false);

			if(!NMtpDeleteFile(path)) {
				int error_code = Marshal.GetLastWin32Error();
				string message = new System.ComponentModel.Win32Exception(error_code).Message;
				if(error_code == ERROR_FILE_NOT_FOUND) {
					throw new FileNotFoundException(message);
				} else {
					throw new IOException(message);
				}
			}
		}

		/// <summary>
		/// ファイルの存在確認
		/// </summary>
		/// <param name="path">ファイル</param>
		/// <returns>true で存在</returns>
		public static bool Exists(string path)
		{
			Directory.CheckInvalidPathName(path, false);

			WIN32_FIND_DATA find_data;

			if(NMtpGetStatus(path, out find_data)) {
				if((find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0) {
					return true;
				}
			}
			return false;
		}

		private const int timeCreation = 0;
		private const int timeLastWrite = 1;
		private const int timeLastAccess = 2;

		private static DateTime GetTime(string path, int time_type, bool local_flag)
		{
			Directory.CheckInvalidPathName(path, false);

			WIN32_FIND_DATA find_data;
			DateTime datetime = new DateTime();

			if(NMtpGetStatus(path, out find_data)) {
				if((find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0) {
					if(time_type == timeLastWrite) {
						datetime = DateTime.FromFileTime((((long)find_data.ftLastWriteTime.dwHighDateTime) << 32) + find_data.ftLastWriteTime.dwLowDateTime);
					} else if(time_type == timeCreation) {
						datetime = DateTime.FromFileTime((((long)find_data.ftCreationTime.dwHighDateTime) << 32) + find_data.ftCreationTime.dwLowDateTime);
					} else if(time_type == timeLastAccess) {
						datetime = DateTime.FromFileTime((((long)find_data.ftLastAccessTime.dwHighDateTime) << 32) + find_data.ftLastAccessTime.dwLowDateTime);
					}
				}
			} else {
				int error_code = Marshal.GetLastWin32Error();
				string message = new System.ComponentModel.Win32Exception(error_code).Message;
				if(error_code == ERROR_PATH_NOT_FOUND) {
					throw new FileNotFoundException(message);
				} else {
					throw new IOException(message);
				}
			}
			if(local_flag) {
				datetime.ToLocalTime();
			}
			return datetime;
		}

		/// <summary>
		/// ファイルの作成日時を取得
		/// </summary>
		/// <param name="path">ファイル</param>
		/// <returns>作成日時</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	ファイルにアクセスできない
		/// </exception>
		public static DateTime GetCreationTime(string path)
		{
			return GetTime(path, timeCreation, true);
		}

		/// <summary>
		/// ファイルの作成日時(世界協定時刻)を取得
		/// </summary>
		/// <param name="path">ファイル</param>
		/// <returns>作成日時(世界協定時刻)</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	ファイルにアクセスできない
		/// </exception>
		public static DateTime GetCreationTimeUtc(string path)
		{
			return GetTime(path, timeCreation, false);
		}

		/// <summary>
		/// ファイルの最後の書込日時を取得
		/// </summary>
		/// <param name="path">ファイル</param>
		/// <returns>書込日時</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	ファイルにアクセスできない
		/// </exception>
		public static DateTime GetLastWriteTime(string path)
		{
			return GetTime(path, timeLastWrite, true);
		}

		/// <summary>
		/// ファイルの最後の書込日時(世界協定時刻)を取得
		/// </summary>
		/// <param name="path">ファイル</param>
		/// <returns>書込日時(世界協定時刻)</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	ファイルにアクセスできない
		/// </exception>
		public static DateTime GetLastWriteTimeUtc(string path)
		{
			return GetTime(path, timeLastWrite, false);
		}

		/// <summary>
		/// ファイルの最後のアクセス日時(世界協定時刻)を取得
		/// </summary>
		/// <param name="path">ファイル</param>
		/// <returns>アクセス日時(世界協定時刻)</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	ファイルにアクセスできない
		/// </exception>
		public static DateTime GetLastAccessTime(string path)
		{
			return GetTime(path, timeLastAccess, true);
		}

		/// <summary>
		/// ファイルの最後のアクセス日時(世界協定時刻)を取得
		/// </summary>
		/// <param name="path">ファイル</param>
		/// <returns>アクセス日時(世界協定時刻)</returns>
		/// <exception cref="ArgumentNullException">
		///	path が null
		///	</exception>
		///	<exception cref="ArgumentException">
		///	path が長さ 0 の文字列か、空白のみか、不正な文字が含まれている
		/// </exception>
		///	<exception cref="PathTooLongException">
		///	path がシステム定義の最大値を超えている
		/// </exception>
		///	<exception cref="DirectoryNotFoundException">
		///	指定されたパスが正しくない
		/// </exception>
		///	<exception cref="IOException">
		///	ファイルにアクセスできない
		/// </exception>
		public static DateTime GetLastAccessTimeUtc(string path)
		{
			return GetTime(path, timeLastAccess, false);
		}

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpCopyFile", CharSet=CharSet.Unicode, SetLastError=true)]
		protected static extern bool NMtpCopyFile(string src, string dst, int fail_exists);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpMoveFile", CharSet=CharSet.Unicode, SetLastError=true)]
		protected static extern bool NMtpMoveFile(string src, string dst);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpDeleteFile", CharSet=CharSet.Unicode, SetLastError=true)]
		protected static extern bool NMtpDeleteFile(string path);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpGetStatus", CharSet=CharSet.Unicode, SetLastError=true)]
		protected static extern bool NMtpGetStatus(string path, out WIN32_FIND_DATA find_data);
	}

	/// <summary>
	/// ファイルを開くダイアログ表示クラス
	/// </summary>
	public class FileDialog {
		/// <summary>
		/// 選択したデバイスパスの先頭の PC/コンピュータ を削除
		/// </summary>
		protected const uint DeleteTopPC = 0x00000001;
		/// <summary>
		/// SaveDialog でファイルの上書き確認を行う
		/// </summary>
		protected const uint SaveOverWritePrompt = 0x00000002;
		/// <summary>
		/// SaveDialog で指定された拡張子のファイルのみ選択可能
		/// </summary>
		protected const uint StrictFileTypes = 0x00000004;
		/// <summary>
		/// カレントフォルダを変更しない
		/// </summary>
		protected const uint NoChangeDir = 0x00000008;
		/// <summary>
		/// 複数項目選択可能
		/// </summary>
		protected const uint AllowMultiSelect = 0x00000200;
		/// <summary>
		/// フォルダの存在が必要
		/// </summary>
		protected const uint PathMustExist = 0x00000800;
		/// <summary>
		/// SaveDialog でファイルの存在が必要
		/// </summary>
		protected const uint FileMustExist = 0x00001000;
		/// <summary>
		/// SaveDialog で存在しない場合通知
		/// </summary>
		protected const uint SaveCreatePrompt = 0x00002000;
		/// <summary>
		/// SaveDialog で読み取り専用不可
		/// </summary>
		protected const uint NoReadOnlyReturn = 0x00008000;
		/// <summary>
		/// 最近開いたアイテムや保存した場所を非表示
		/// </summary>
		protected const uint HideMruPlace = 0x00020000;
		/// <summary>
		/// ナビゲーションペインを非表示
		/// </summary>
		protected const uint HidePinnedPlaces = 0x00040000;
		/// <summary>
		/// ショートカットを対象としない
		/// </summary>
		protected const uint NoDereferenceLinks = 0x00100000;
		/// <summary>
		/// 最新のドキュメントに開いたファイルを追加しない
		/// </summary>
		protected const uint DontAddToRecent = 0x02000000;
		/// <summary>
		/// 隠しファイルとシステムアイテムを含む
		/// </summary>
		protected const uint ShowHidden = 0x10000000;
		/// <summary>
		/// OpenFile でプレビューを表示する
		/// </summary>
		protected const uint PreviewPanelOn = 0x40000000;

		protected const int MAX_PATH = 260;

		/// <summary>
		/// <c>FileDialog</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		public FileDialog() {
			Handle = IntPtr.Zero;
			CheckPathExists = true;
			CheckFileExists = true;
			RestoreDirectory = true;
			FilterIndex = 1;
		}
		/// <summary>
		/// <c>FileDialog</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		/// <param name="hwnd">親ウインドウハンドル</param>
		public FileDialog(IntPtr hwnd) {
            Handle = hwnd;
            CheckPathExists = true;
            CheckFileExists = true;
            RestoreDirectory = true;
            FilterIndex = 1; 
		}

		protected virtual uint MakeFolderOption()
		{
			uint option = 0;
			if(CheckFileExists) {
				option |= FileMustExist;
			}
			if(CheckPathExists) {
				option |= PathMustExist;
			}
			if(RestoreDirectory) {
				option |= NoChangeDir;
			}
			return option;
		}

		/// <summary>
		/// 既定の拡張子を取得または設定
		/// </summary>
		public string DefaultExt { get; set; }
		/// <summary>
		/// 既定のファイル名を設定、選択されたファイル名を取得
		/// </summary>
		public string FileName { get; set; }
		/// <summary>
		/// フィルター文字列の取得または設定
		/// </summary>
		public string Filter { get; set; }
		/// <summary>
		/// 既定フィルター番号の取得又は設定
		/// </summary>
		public int FilterIndex { get; set; }
		/// <summary>
		/// <para>既定フォルダー名の取得または設定</para>
		/// <para>※MTP デバイス上のパスは指定できません</para>
		/// </summary>
		public string InitialDirectory { get; set; }
		/// <summary>
		/// ダイアログタイトルの取得または設定
		/// </summary>
		public string Title { get; set; }
		/// <summary>
		/// 親ウインドウハンドルの取得または設定
		/// </summary>
		public IntPtr Handle { get; set; }
		/// <summary>
		/// 存在しないファイルの名前をユーザーが指定した場合に警告を表示するかどうかを示す値を取得または設定
		/// </summary>
		public bool CheckFileExists { get; set; }
		/// <summary>
		/// 存在しないパスをユーザーが指定した場合に警告を表示するかどうかを示す値を取得または設定
		/// </summary>
		public bool CheckPathExists { get; set; }
		/// <summary>
		/// カレントフォルダを変更するかどうかを示す値を取得または設定
		/// </summary>
		public bool RestoreDirectory { get; set; }
	}

	/// <summary>
	/// ファイルを開くダイアログ表示クラス
	/// </summary>
	public class OpenFileDialog : FileDialog {
		protected List<string> _file_list = new List<string>();

		/// <summary>
		/// <c>OpenFileDialog</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		public OpenFileDialog() : base() {
			AllowSelect = false;
			PreviewPanel = false;
		}
		/// <summary>
		/// <c>OpenFileDialog</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		/// <param name="hwnd">親ウインドウハンドル</param>
		public OpenFileDialog(IntPtr hwnd) : base(hwnd) {
            AllowSelect = false;
            PreviewPanel = false;
		}

		/// <summary>
		/// 複数ファイルを選択可能かどうかを示す値を取得または設定
		/// </summary>
		public bool AllowSelect { get; set; }
		/// <summary>
		/// プレビュー表示をするかどうかを示す値を取得または設定
		/// </summary>
		public bool PreviewPanel { get; set; }
		/// <summary>
		/// 選択された全てのファイル名を取得
		/// </summary>
		public string [] FileNames {
			get {
				return _file_list.ToArray();
			}
		}

		protected override uint MakeFolderOption()
		{
			uint option = base.MakeFolderOption();
			if(AllowSelect) {
				option |= AllowMultiSelect;
			}
			if(PreviewPanel) {
				option |= PreviewPanelOn;
			}
			return option;
		}

		/// <summary>
		/// ファイルを開くダイアログを表示
		/// </summary>
		/// <returns>DialogResult.OK でファイル選択</returns>
		public System.Windows.Forms.DialogResult ShowDialog()
		{
			StringBuilder sname = new StringBuilder(MAX_PATH);
			int result;
            uint option = MakeFolderOption();
			result = NMtpOpenDialog(Handle, Title, InitialDirectory, FileName, Filter, FilterIndex, DefaultExt, sname, MAX_PATH, option);
			if(result != 0) {
				FileName = sname.ToString();
				if((option & AllowMultiSelect) == AllowMultiSelect) {
					_file_list = new List<string>();
					_file_list.Add(FileName);
					while(NMtpOpenDialogNext(result, sname, MAX_PATH)) {
						_file_list.Add(sname.ToString());
					}
					NMtpOpenDialogClose(result);
				}
				return System.Windows.Forms.DialogResult.OK;
			}
			return System.Windows.Forms.DialogResult.Cancel;
		}

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpOpenDialog", CharSet=CharSet.Unicode)]
		protected static extern int NMtpOpenDialog(IntPtr wnd, string title, string folder, string name, string filter, int index, string ext, StringBuilder filename, uint length, uint flag);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpOpenDialogNext", CharSet=CharSet.Unicode)]
		protected static extern bool NMtpOpenDialogNext(int handle, StringBuilder filename, uint length);

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpOpenDialogClose", CharSet=CharSet.Unicode)]
		protected static extern bool NMtpOpenDialogClose(int handle);
	}
	/// <summary>
	/// ファイル保存ダイアログ表示クラス
	/// </summary>
	public class SaveFileDialog : FileDialog {
		/// <summary>
		/// <c>SaveFileDialog</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		public SaveFileDialog() : base() {
			CreatePrompt = true;
		}
		/// <summary>
		/// <c>SaveFileDialog</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		/// <param name="hwnd">親ウインドウハンドル</param>
		public SaveFileDialog(IntPtr hwnd) : base(hwnd) {
			CreatePrompt = true;
		}

		/// <summary>
		/// 存在しないファイルが選択された場合に許可表示を行うかどうかを示す値を取得または設定
		/// </summary>
		public bool CreatePrompt { get; set; }

		protected override uint MakeFolderOption()
		{
			uint option = base.MakeFolderOption();
			if(CreatePrompt) {
				option |= SaveCreatePrompt;
			}
			return option;
		}
		/// <summary>
		/// ファイルを保存ダイアログを表示
		/// </summary>
		/// <returns>DialogResult.OK でファイル選択</returns>
		public System.Windows.Forms.DialogResult ShowDialog()
		{
			StringBuilder sname = new StringBuilder(MAX_PATH);
			if(NMtpSaveDialog(Handle, Title, InitialDirectory, FileName, Filter, FilterIndex, DefaultExt, sname, MAX_PATH, MakeFolderOption())) {
				FileName = sname.ToString();
				return System.Windows.Forms.DialogResult.OK;
			}
			return System.Windows.Forms.DialogResult.Cancel;
		}

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpSaveDialog", CharSet=CharSet.Unicode)]
		protected static extern bool NMtpSaveDialog(IntPtr wnd, string title, string folder, string name, string filter, int filter_index, string ext, StringBuilder filename, int length, uint flag);
	}
	/// <summary>
	/// フォルダ選択ダイアログ表示クラス
	/// </summary>
	public class FolderBrowserDialog {
		private const int MAX_DIR_PATH = 248;

		/// <summary>
		/// <c>FolderBrowserDialog</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		public FolderBrowserDialog() {
			Handle = IntPtr.Zero;
		}
		/// <summary>
		/// <c>FolderBrowserDialog</c>クラスの新規インスタンスを初期化します。
		/// </summary>
		/// <param name="hwnd">親ウインドウハンドル</param>
		public FolderBrowserDialog(IntPtr hwnd) {
			Handle = hwnd;
		}

		/// <summary>
		/// 親ウインドウハンドルの取得または設定
		/// </summary>
		public IntPtr Handle { get; set; }
		/// <summary>
		/// ダイアログタイトルの取得または設定
		/// </summary>
		public string Description { get; set; }
		/// <summary>
		/// <para>既定フォルダー名の取得または設定</para>
		/// <para>※MTP デバイス上のパスは指定できません</para>
		/// </summary>
		public string RootFolder { get; set; }
		/// <summary>
		/// 選択されたフォルダ名を取得
		/// </summary>
		public string SelectPath { get; set; }

		/// <summary>
		/// フォルダ選択ダイアログを表示
		/// </summary>
		/// <returns>DialogResult.OK でフォルダ選択</returns>
		public System.Windows.Forms.DialogResult ShowDialog()
		{
			StringBuilder sname = new StringBuilder(MAX_DIR_PATH);
			if(NMtpOpenFolderDialog(Handle, Description, RootFolder, sname, MAX_DIR_PATH, 0)) {
				SelectPath = sname.ToString();
				return System.Windows.Forms.DialogResult.OK;
			}
			return System.Windows.Forms.DialogResult.Cancel;
		}

		[DllImport("nMtpUtil.DLL", EntryPoint="NMtpOpenFolderDialog", CharSet=CharSet.Unicode)]
		protected static extern bool NMtpOpenFolderDialog(IntPtr wnd, string title, string folder, StringBuilder filename, int length, int flag);
	}
}
